""" This module contains pre-configurations for finetuning on the demo datasets. """
from dataclasses import dataclass
from typing import Optional, Tuple

from docarray import DocumentArray

from now.constants import Apps, DatasetTypes
from now.now_dataclasses import UserInput

DEFAULT_EPOCHS = 50
DEFAULT_HIDDEN_SIZES = (128,)
DEFAULT_NUM_VAL_QUERIES = 50
DEFAULT_FINETUNED_EMBEDDING_SIZE = 128
DEFAULT_BATCH_SIZE = 128
DEFAULT_TRAIN_VAL_SPLIT_RATIO = 0.9
DEFAULT_EVAL_MATCH_LIMIT = 20
DEFAULT_NUM_ITEMS_PER_CLASS = 4
DEFAULT_LEARNING_RATE = 5e-4
DEFAULT_EARLY_STOPPING_PATIENCE = 5
DEFAULT_POS_MINING_START = 'hard'
DEFAULT_NEG_MINING_START = 'hard'


@dataclass
class FinetuneSettings:
    perform_finetuning: bool
    model_name: str
    add_embeddings: bool
    bi_modal: bool  # atm, bi-modal means text and some blob value
    loss: str
    pre_trained_embedding_size: Optional[int] = None
    finetuned_model_artifact: Optional[str] = None
    token: Optional[str] = None

    hidden_sizes: Tuple[int] = DEFAULT_HIDDEN_SIZES
    batch_size: int = DEFAULT_BATCH_SIZE
    epochs: int = DEFAULT_EPOCHS
    finetune_layer_size: int = DEFAULT_FINETUNED_EMBEDDING_SIZE
    train_val_split_ration: int = DEFAULT_TRAIN_VAL_SPLIT_RATIO
    num_val_queries: int = DEFAULT_NUM_VAL_QUERIES
    eval_match_limit: int = DEFAULT_EVAL_MATCH_LIMIT
    num_items_per_class: int = DEFAULT_NUM_ITEMS_PER_CLASS
    learning_rate: int = DEFAULT_LEARNING_RATE
    pos_mining_strat: str = DEFAULT_POS_MINING_START
    neg_mining_strat: str = DEFAULT_NEG_MINING_START
    early_stopping_patience: int = DEFAULT_EARLY_STOPPING_PATIENCE


def _is_finetuning(
    user_input: UserInput, dataset: DocumentArray, finetuneable_datasets: Tuple
) -> bool:
    return False
    # if (
    #     (
    #         user_input.dataset_type == DatasetTypes.DEMO
    #         and user_input.dataset_name in finetuneable_datasets
    #     )
    #     or (
    #         user_input.dataset_type != DatasetTypes.DEMO
    #         and all(['finetuner_label' in d.tags for d in dataset])
    #     )
    #     # TODO: enable finetuning again when FinetunerExecutor has been updated
    #     # or user_input.app_instance.app_name == Apps.TEXT_TO_TEXT_AND_IMAGE
    # ):
    #     return True
    # else:
    #     return False


def _is_bi_modal(user_input: UserInput, dataset: DocumentArray) -> bool:
    if user_input.app_instance.app_name == Apps.TEXT_TO_TEXT_AND_IMAGE:
        return False
    elif user_input.dataset_type != DatasetTypes.DEMO:
        has_blob = any([d.blob != b'' for d in dataset])
        has_text = any([d.text != '' for d in dataset])
        return has_text and has_blob
    else:
        return True  # right now all demo cases are bi-modal


def parse_finetune_settings(
    user_input: UserInput,
    dataset: DocumentArray,
    model_name: str,
    loss: str,
    finetune_datasets: Tuple = (),
    add_embeddings: bool = True,
    pre_trained_embedding_size: Optional[int] = None,
) -> FinetuneSettings:
    """This function parses the user input configuration into the finetune settings"""
    return FinetuneSettings(
        perform_finetuning=_is_finetuning(user_input, dataset, finetune_datasets),
        bi_modal=_is_bi_modal(user_input, dataset),
        pre_trained_embedding_size=pre_trained_embedding_size,
        model_name=model_name,
        loss=loss,
        add_embeddings=add_embeddings,
    )



from jina import Client, Document

flow_id = ''
api_key = ''
search_text = ''
limit = 60


client = Client(host=f'grpcs://nowapi-{flow_id}.wolf.jina.ai')


def call():
    result = client.post(
        '/search',
        inputs=Document(chunks=Document(text=search_text)),
        parameters={
            'api_key': api_key,
            'limit': limit,
        },
    )
    return result[0].matches[:, 'uri']


for uri in call():
    print(uri)



import os
from typing import Dict, List, Union

from docarray import Document, DocumentArray

from now.constants import Modalities


def _get_modality(document: Document):
    """
    Detect document's modality based on its `modality` or `mime_type` attributes.
    """
    for modality in Modalities():
        if modality in document.modality or modality in document.mime_type:
            return modality
    return None


def _get_multi_modal_format(document: Document) -> Document:
    """
    Create a multimodal docarray structure from a unimodal `Document`.
    """
    from now.app.text_to_video.app import TextToVideo

    modality = _get_modality(document)
    if document.blob:
        new_doc = Document(chunks=[Document(blob=document.blob)])
    elif document.uri:
        file_type = os.path.splitext(document.uri)[-1].replace('.', '')
        new_doc = Document(chunks=[Document(uri=document.uri)])
        if file_type in TextToVideo().supported_file_types:
            modality = Modalities.VIDEO
        else:
            modality = Modalities.IMAGE
    elif document.text:
        new_doc = Document(chunks=[Document(text=document.text)])
        modality = Modalities.TEXT
    else:
        raise Exception(f'Document {document} cannot be transformed.')
    if modality:
        new_doc.chunks[0].modality = modality
    return new_doc


def transform_uni_modal_data(documents: DocumentArray) -> DocumentArray:
    """
    Transform unimodal `Documents` into standardized format, which looks like this:
    Document(
        tags={'color': 'red', 'author': 'me'},
        chunks=[
            Document(
                text='jina ai', tags={'color': 'red', 'author': 'me'}
            )
        ],
    )
    """
    transformed_docs = DocumentArray()
    for document in documents:
        new_doc = _get_multi_modal_format(document)
        new_doc.chunks[0].tags['field_name'] = 'default_field'
        new_doc.chunks[0].embedding = document.embedding
        new_doc.tags = document.tags
        new_doc.chunks[0].tags.update(document.tags)
        new_doc.chunks[0].mime_type = new_doc.chunks[0].modality
        transformed_docs.append(new_doc)

    return transformed_docs


def transform_multi_modal_data(
    documents: DocumentArray, field_names: Dict[int, str], search_fields: List[str]
):
    """
    Transforms multimodal data into standardized format, which looks like this:
    Document(
        tags={'color': 'red', 'author': 'me'},
        chunks=[
            Document(
                text='jina ai', tags={'color': 'red', 'author': 'me'}
            ),
            Document(
                uri='pic.jpg', tags={'color': 'red', 'author': 'me'}
            ),
        ],
    )
    """
    for document in documents:
        new_chunks = []
        for position, chunk in enumerate(document.chunks):
            field_name = field_names[position]
            modality = chunk.modality or _get_modality(chunk)
            if field_name in search_fields:
                new_chunks.append(
                    Document(
                        content=chunk.content,
                        uri=chunk.uri,
                        modality=modality,
                        mime_type=modality,
                        tags={'field_name': field_name},
                    )
                )
            else:
                document.tags[field_name] = chunk.content
        document.chunks = new_chunks
        for chunk in document.chunks:
            chunk.tags.update(document.tags)

    return documents


def transform_docarray(
    documents: Union[Document, DocumentArray],
    search_fields: List[str],
) -> DocumentArray:
    """
    Gets either multimodal or unimodal data and turns it into standardized format.

    :param documents: Data to be transformed.
    :param search_fields: Field names for neural search. Only required if multimodal data is given.
    """
    if documents and documents[0].chunks:
        if 'multi_modal_schema' not in documents[0]._metadata:
            raise RuntimeError(
                'Multi-modal schema is not provided. Please prepare your data following this guide - '
                'https://docarray.jina.ai/datatypes/multimodal/'
            )
        field_names = {
            int(field_info['position']): field_name
            for field_name, field_info in documents[0]
            ._metadata['multi_modal_schema']
            .items()
        }
        documents = transform_multi_modal_data(
            documents=documents, field_names=field_names, search_fields=search_fields
        )
    else:
        documents = transform_uni_modal_data(documents=documents)
    return documents



# -*- coding: utf-8 -*-
"""
provide colorized output
"""
import sys

from prompt_toolkit.shortcuts import print_tokens


def _print_token_factory(col):
    """Internal helper to provide color names."""

    def _helper(msg):
        tokens = [('fg:' + col, msg)]
        print_tokens(tokens)

    def _helper_no_terminal(msg):
        # workaround if we have no terminal
        print(msg)

    if sys.stdout.isatty():
        return _helper
    else:
        return _helper_no_terminal


# used this for color source:
# http://unix.stackexchange.com/questions/105568/how-can-i-list-the-available-color-names
yellow = _print_token_factory('#dfaf00')
blue = _print_token_factory('#0087ff')
gray = _print_token_factory('#6c6c6c')

# TODO
# black
# red
# green
# magenta
# cyan
# white



import os

import hubble


def get_default_request_body(deployment_type, secured, remote_host=None):
    request_body = {}
    if deployment_type == 'local':
        request_body['host'] = (
            'localhost' if os.environ.get('NOW_TESTING', False) else 'gateway'
        )
        request_body['port'] = 8080
    elif deployment_type == 'remote':
        if remote_host:
            request_body['host'] = remote_host
        else:
            raise ValueError('Remote host must be provided for remote deployment')
    if secured:
        if 'WOLF_TOKEN' in os.environ:
            os.environ['JINA_AUTH_TOKEN'] = os.environ['WOLF_TOKEN']
        request_body['jwt'] = {'token': hubble.get_token()}
    return request_body



from __future__ import annotations, print_function, unicode_literals

import base64
import json
import os
import shutil
import signal
import sys
import tempfile
from collections.abc import MutableMapping
from concurrent.futures import ThreadPoolExecutor
from os.path import expanduser as user
from typing import Dict, List, Optional, Union

import boto3
import hubble
import yaml
from docarray import Document, DocumentArray
from jina.jaml import JAML
from pyfiglet import Figlet

from now.thirdparty.PyInquirer.prompt import prompt


def download_file(path, r_raw):
    with path.open("wb") as f:
        shutil.copyfileobj(r_raw, f)


def download(url, filename):
    import functools
    import pathlib

    import requests
    from tqdm.auto import tqdm

    r = requests.get(url, stream=True, allow_redirects=True)
    if r.status_code != 200:
        r.raise_for_status()  # Will only raise for 4xx codes, so...
        raise RuntimeError(f"Request to {url} returned status code {r.status_code}")
    file_size = int(r.headers.get('Content-Length', 0))

    path = pathlib.Path(filename).expanduser().resolve()
    path.parent.mkdir(parents=True, exist_ok=True)

    desc = "(Unknown total file size)" if file_size == 0 else ""
    r.raw.read = functools.partial(
        r.raw.read, decode_content=True
    )  # Decompress if needed

    if any(map(lambda x: x in os.environ, ['NOW_CI_RUN', 'NOW_EXAMPLES'])):
        download_file(path, r.raw)
    else:
        with tqdm.wrapattr(r.raw, "read", total=file_size, desc=desc) as r_raw:
            download_file(path, r_raw)

    return path


def my_handler(signum, frame, spinner):
    with spinner.hidden():
        sys.stdout.write("Program terminated!\n")
    spinner.stop()
    exit(0)


def flow_definition(dirpath) -> Dict:
    with open(dirpath) as f:
        return yaml.safe_load(f.read())


class BetterEnum:
    def __iter__(self):
        return [getattr(self, x) for x in dir(self) if ('__' not in x)].__iter__()


def to_camel_case(text):
    """Convert text to camel case in great coding style"""
    s = text.replace("_", " ")
    s = s.split()
    return ''.join(i.capitalize() for i in s)


sigmap = {signal.SIGINT: my_handler, signal.SIGTERM: my_handler}


class EnvironmentVariables:
    def __init__(self, envs: Dict):
        self._env_keys_added: Dict = envs

    def __enter__(self):
        for key, val in self._env_keys_added.items():
            os.environ[key] = str(val)

    def __exit__(self, *args, **kwargs):
        for key in self._env_keys_added.keys():
            os.unsetenv(key)


def add_env_variables_to_flow(app_instance, env_dict: Dict):
    with EnvironmentVariables(env_dict):
        app_instance.flow_yaml = JAML.expand_dict(app_instance.flow_yaml, env_dict)


def write_env_file(env_file, config):
    config_string = '\n'.join([f'{key}={value}' for key, value in config.items()])
    with open(env_file, 'w+') as fp:
        fp.write(config_string)


def write_flow_file(flow_yaml_content, new_yaml_file_path):
    with open(new_yaml_file_path, 'w') as f:
        JAML.dump(
            flow_yaml_content,
            f,
            indent=2,
            allow_unicode=True,
            Dumper=Dumper,
        )


@hubble.login_required
def jina_auth_login():
    pass


def get_info_hubble(user_input):
    with open(user('~/.jina/config.json')) as fp:
        config_val = json.load(fp)
        user_token = config_val['auth_token']
    client = hubble.Client(token=user_token, max_retries=None, jsonify=True)
    response = client.get_user_info()
    user_input.admin_emails = (
        [response['data']['email']] if 'email' in response['data'] else []
    )
    if not user_input.admin_emails:
        print(
            'Your hubble account is not verified. Please verify your account to deploy your flow as admin.'
        )
    user_input.jwt = {'token': user_token}
    return response['data'], user_token


def print_headline():
    f = Figlet(font='slant')
    print('Welcome to:')
    print(f.renderText('Jina NOW'))
    print('Get your search case up and running - end to end.\n')
    print(
        'You can choose between image and text search. \nJina NOW trains a model, pushes it to Jina AI Cloud '
        'and deploys a Flow and a playground app in the cloud or locally. \nCheck out one of our demos or bring '
        'your own data.\n'
    )
    print('If you want learn more about our framework please visit docs.jina.ai')
    print(
        '💡 Make sure you give enough memory to your Docker daemon. '
        '5GB - 8GB should be okay.'
    )
    print()


def maybe_prompt_user(questions, attribute, **kwargs):
    """
    Checks the `kwargs` for the `attribute` name. If present, the value is returned directly.
    If not, the user is prompted via the cmd-line using the `questions` argument.

    :param questions: A dictionary that is passed to `PyInquirer.prompt`
        See docs: https://github.com/CITGuru/PyInquirer#documentation
    :param attribute: Name of the value to get. Make sure this matches the name in `kwargs`

    :return: A single value of either from `kwargs` or the user cli input.
    """
    if kwargs and attribute in kwargs:
        return kwargs[attribute]
    else:
        answer = prompt(questions)
        return answer[attribute]


def prompt_value(
    name: str,
    prompt_message: str,
    prompt_type: str = 'input',
    choices: Optional[List[Union[Dict, str]]] = None,
    **kwargs: Dict,
):
    qs = {'name': name, 'type': prompt_type, 'message': prompt_message}

    if choices is not None:
        qs['choices'] = choices
        qs['type'] = 'list'
    return maybe_prompt_user(qs, name, **kwargs)


def get_local_path(tmpdir, path_s3):
    return os.path.join(
        str(tmpdir),
        base64.b64encode(bytes(path_s3, "utf-8")).decode("utf-8"),
    )


def download_from_bucket(tmpdir, uri, bucket):
    path_s3 = '/'.join(uri.split('/')[3:])
    path_local = get_local_path(tmpdir, path_s3)
    bucket.download_file(
        path_s3,
        path_local,
    )
    return path_local


def convert_fn(
    d: Document, tmpdir, aws_access_key_id, aws_secret_access_key, aws_region_name
) -> Document:
    """Downloads files and tags from S3 bucket and updates the content uri and the tags uri to the local path"""
    bucket = get_bucket(
        uri=d.uri,
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
        region_name=aws_region_name,
    )
    d.tags['uri'] = d.uri

    d.uri = download_from_bucket(tmpdir, d.uri, bucket)
    if 'tag_uri' in d.tags:
        local_tag_uri = download_from_bucket(tmpdir, d.tags['tag_uri'], bucket)
        with open(local_tag_uri, 'r') as fp:
            tags = json.load(fp)
            tags = flatten_dict(tags)
            d.tags.update(tags)
        del d.tags['tag_uri']
    return d


def get_bucket(uri, aws_access_key_id, aws_secret_access_key, region_name):
    session = boto3.session.Session(
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
        region_name=region_name,
    )
    bucket = session.resource('s3').Bucket(uri.split('/')[2])
    return bucket


def maybe_download_from_s3(
    docs: DocumentArray, tmpdir: tempfile.TemporaryDirectory, user_input, max_workers
):
    """Downloads file to local temporary dictionary, saves S3 URI to `tags['uri']` and modifies `uri` attribute of
    document to local path in-place.

    :param doc: document containing URI pointing to the location on S3 bucket
    :param tmpdir: temporary directory in which files will be saved
    :param user_input: User iput which contain aws credentials
    :param max_workers: number of threads to create in the threadpool executor to make execution faster
    """

    flat_docs = docs['@c']
    filtered_docs = [c for c in flat_docs if c.uri.startswith('s3://')]

    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        futures = []
        for c in filtered_docs:
            f = executor.submit(
                convert_fn,
                c,
                tmpdir,
                user_input.aws_access_key_id,
                user_input.aws_secret_access_key,
                user_input.aws_region_name,
            )
            futures.append(f)
        for f in futures:
            f.result()


def flatten_dict(d, parent_key='', sep='_'):
    items = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, MutableMapping):
            items.extend(flatten_dict(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)


def _get_context_names(contexts, active_context=None):
    names = [c for c in contexts] if contexts is not None else []
    if active_context is not None:
        names.remove(active_context)
        names = [active_context] + names
    return names


def get_flow_id(host):
    return host.split('.wolf.jina.ai')[0].split('grpcs://')[-1]


class Dumper(yaml.Dumper):
    def increase_indent(self, flow=False, *args, **kwargs):
        return super().increase_indent(flow=flow, indentless=False)



import base64
from itertools import zip_longest
from typing import List

from docarray import Document, DocumentArray
from fastapi import APIRouter

from deployment.bff.app.v1.models.text import NowTextSearchRequestModel
from deployment.bff.app.v1.models.text_and_image import (
    NowTextImageIndexRequestModel,
    NowTextImageResponseModel,
    NowTextImageSearchRequestModel,
)
from deployment.bff.app.v1.routers.helper import jina_client_post, process_query

router = APIRouter()


# Index
@router.post(
    "/index",
    summary='Add more image or text data to the indexer',
)
def index(data: NowTextImageIndexRequestModel):
    """
    Append the list of image or text data to the indexer. Each image data should be
    `base64` encoded using human-readable characters - `utf-8`.
    """
    index_docs = DocumentArray()
    for text, image, uri, tags in zip_longest(
        data.texts, data.images, data.uris, data.tags
    ):
        if bool(image) + bool(uri) + bool(text) > 1:
            raise ValueError(
                f'Can only set one value but have image={image}, uri={uri}, text={text}'
            )
        if image:
            base64_bytes = image.encode('utf-8')
            image = base64.decodebytes(base64_bytes)
            index_docs.append(Document(blob=image, tags=tags))
        elif text:
            index_docs.append(Document(text=text, tags=tags))
        else:
            index_docs.append(Document(uri=uri, tags=tags))

    jina_client_post(
        data=data,
        inputs=index_docs,
        endpoint='/index',
    )


# Search
@router.post(
    "/search",
    response_model=List[NowTextImageResponseModel],
    summary='Search image or text data via image or text as query',
)
def search(data: NowTextImageSearchRequestModel):
    """
    Retrieve matching images or texts for a given image or text query. Image query should be
    `base64` encoded using human-readable characters - `utf-8`.
    """
    # enable text search as well
    query_doc, filter_query = process_query(
        blob=data.image, text=data.text, uri=data.uri, conditions=data.filters
    )

    docs = jina_client_post(
        data=data,
        inputs=query_doc,
        parameters={'limit': data.limit, 'filter': filter_query},
        endpoint='/search',
    )

    return docs[0].matches.to_dict()


@router.post(
    "/suggestion",
    summary='Get auto complete suggestion for query',
)
def suggestion(data: NowTextSearchRequestModel):
    """
    Return text suggestions for the rest of the query text
    """
    query_doc, filter_query = process_query(
        text=data.text, uri=data.uri, conditions=data.filters
    )

    docs = jina_client_post(
        data=data,
        inputs=query_doc,
        endpoint='/suggestion',
        parameters={'limit': data.limit, 'filter': filter_query, 'apply_bm25': True},
    )
    return docs.to_dict()



import os

from now.constants import EXECUTOR_PREFIX


def handle_test_mode(config):
    if os.environ.get('NOW_TESTING', False):
        from now.executor.autocomplete import NOWAutoCompleteExecutor2
        from now.executor.indexer.in_memory import InMemoryIndexer
        from now.executor.preprocessor import NOWPreprocessor

        # this is a hack to make sure the import is not removed
        if NOWPreprocessor and NOWAutoCompleteExecutor2 and InMemoryIndexer:
            pass

        for k, v in config.items():
            # if 'Encoder' in str(v):
            #     config[k] = config[k].replace('+docker', '')
            if 'Indexer' in str(v):
                config[k] = 'InMemoryIndexer'
            if (
                isinstance(v, str)
                and 'jinahub' in v
                and (
                    # TODO: local testing on Qdrant needs to be disabled. At the moment, Qdrant does not start outside of docker
                    # TODO: same for elastic
                    not 'NOWQdrantIndexer' in v
                    and not 'NOWElasticIndexer' in v
                    and not 'CLIPOnnxEncoder' in v
                    and not 'NOWOCRDetector9' in v
                    and not 'TransformerSentenceEncoder' in v
                )
            ):
                config[k] = config[k].replace(EXECUTOR_PREFIX, '').split('/')[0]



from concurrent.futures import ProcessPoolExecutor
from time import time

from now.client import Client
from now.constants import Apps


def call(client, use_bff):
    if use_bff:
        send_fn = client.send_request_bff
        end_point = 'text-to-video/search'
    else:
        send_fn = client.send_request
        end_point = '/search'
    start = time()

    try:
        send_fn(endpoint=end_point, text='girl on a motorbike', limit=60)
    except Exception as e:
        import traceback

        traceback.print_exc()
    dif = time() - start
    return dif


def benchmark_deployment(jcloud_id, api_key):
    print(f"benchmark flow: {jcloud_id}")

    client = Client(
        jcloud_id=jcloud_id,
        app=Apps.TEXT_TO_VIDEO,
        api_key=api_key,
    )

    for call_point, use_bff in zip(['BFF', 'Gateway'], [True, False]):
        # Latency Test
        start = time()
        for i in range(10):
            call(client=client, use_bff=use_bff)
        print(f'Latency {call_point}: {(time() - start) / 10}s')

        num_queries = 500
        worker = 30
        with ProcessPoolExecutor(max_workers=worker) as executor:
            # QPS test
            start = time()
            futures = []
            latencies = []

            for i in range(num_queries):
                future = executor.submit(call, client, use_bff)
                futures.append(future)
            for future in futures:
                latencies.append(future.result())
            print(f'QPS {call_point}: {num_queries / (time() - start)}s')

        latencies = sorted(latencies)
        for p in [0, 50, 75, 85, 90, 95, 99, 99.9]:
            print(f"P{p} {call_point}: {latencies[int(len(latencies) * p / 100)]}")

        print('\n')


if __name__ == '__main__':
    benchmark_deployment(jcloud_id='jcloud_id', api_key='api_key')



import base64
import os
from typing import Dict, List

from docarray import Document, DocumentArray
from jina.serve.runtimes.gateway.http.models import JinaRequestModel, JinaResponseModel
from pydantic import BaseModel

from deployment.bff.app.v1.models.text import NowTextSearchRequestModel
from deployment.bff.app.v1.models.video import (
    NowVideoIndexRequestModel,
    NowVideoListResponseModel,
    NowVideoResponseModel,
)
from now.app.base.app import JinaNOWApp
from now.common.utils import (
    _get_clip_apps_with_dict,
    common_setup,
    get_email,
    get_indexer_config,
)
from now.constants import CLIP_USES, Apps, Modalities
from now.now_dataclasses import UserInput

NUM_FRAMES_SAMPLED = 3


class TextToVideo(JinaNOWApp):
    def __init__(self):
        super().__init__()

    @property
    def app_name(self) -> str:
        return Apps.TEXT_TO_VIDEO

    @property
    def is_enabled(self) -> bool:
        return True

    @property
    def description(self) -> str:
        return 'Text to video search app'

    @property
    def input_modality(self) -> List[Modalities]:
        return [Modalities.TEXT]

    @property
    def output_modality(self) -> List[Modalities]:
        return [Modalities.VIDEO]

    @property
    def required_docker_memory_in_gb(self) -> int:
        return 12

    def set_flow_yaml(self, **kwargs):
        finetuning = kwargs.get('finetuning', False)
        dataset_len = kwargs.get('dataset_len', 0) * NUM_FRAMES_SAMPLED
        is_jina_email = get_email().split('@')[-1] == 'jina.ai'

        flow_dir = os.path.abspath(os.path.join(__file__, '..'))

        if finetuning:
            if dataset_len > 200_000 and is_jina_email:
                print(f"🚀🚀🚀 You are using high performance flow")
                self.flow_yaml = os.path.join(
                    flow_dir, 'ft-flow-video-clip-high-performance.yml'
                )
            else:
                self.flow_yaml = os.path.join(flow_dir, 'ft-flow-video-clip.yml')
        else:
            if dataset_len > 200_000 and is_jina_email:
                print(f"🚀🚀🚀 You are using high performance flow")
                self.flow_yaml = os.path.join(
                    flow_dir, 'flow-video-clip-high-performance.yml'
                )
            else:
                self.flow_yaml = os.path.join(flow_dir, 'flow-video-clip.yml')

    def setup(
        self, dataset: DocumentArray, user_input: UserInput, kubectl_path
    ) -> Dict:
        indexer_config = get_indexer_config(len(dataset) * NUM_FRAMES_SAMPLED)
        encoder_with = _get_clip_apps_with_dict(user_input)
        env_dict = common_setup(
            app_instance=self,
            user_input=user_input,
            dataset=dataset,
            encoder_uses=CLIP_USES[user_input.deployment_type][0],
            encoder_with=encoder_with,
            encoder_uses_with={
                'pretrained_model_name_or_path': CLIP_USES[user_input.deployment_type][
                    1
                ]
            },
            pre_trained_embedding_size=CLIP_USES[user_input.deployment_type][2],
            indexer_uses=indexer_config['indexer_uses'],
            kubectl_path=kubectl_path,
            indexer_resources=indexer_config['indexer_resources'],
        )
        super().setup(dataset, user_input, kubectl_path)
        return env_dict

    @property
    def bff_mapping_fns(self):
        def search_text_to_video_request_mapping_fn(
            request: NowTextSearchRequestModel,
        ) -> JinaRequestModel:
            jina_request_model = JinaRequestModel()
            jina_request_model.data = [Document(chunks=[Document(text=request.text)])]
            jina_request_model.parameters = {
                'limit': request.limit,
                'api_key': request.api_key,
                'jwt': request.jwt,
            }
            return jina_request_model

        def search_video_response_mapping_fn(
            request: NowTextSearchRequestModel, response: JinaResponseModel
        ) -> List[NowVideoResponseModel]:
            docs = response.data
            limit = request.limit
            return docs[0].matches[:limit].to_dict()

        def index_text_to_video_request_mapping_fn(
            request: NowVideoIndexRequestModel,
        ) -> JinaRequestModel:
            index_docs = DocumentArray()
            for video, uri, tags in zip(request.videos, request.uris, request.tags):
                if bool(video) + bool(uri) != 1:
                    raise ValueError(
                        f'Can only set one value but have video={video}, uri={uri}'
                    )
                if video:
                    base64_bytes = video.encode('utf-8')
                    message = base64.decodebytes(base64_bytes)
                    index_docs.append(Document(blob=message, tags=tags))
                else:
                    index_docs.append(Document(uri=uri, tags=tags))
            return JinaRequestModel(data=index_docs)

        def no_response_mapping_fn(_: JinaResponseModel) -> BaseModel:
            return BaseModel()

        return {
            '/search': (
                NowTextSearchRequestModel,
                NowVideoListResponseModel,
                search_text_to_video_request_mapping_fn,
                search_video_response_mapping_fn,
            ),
            '/index': (
                NowVideoIndexRequestModel,
                BaseModel,
                index_text_to_video_request_mapping_fn,
                no_response_mapping_fn,
            ),
        }

    @property
    def max_request_size(self) -> int:
        """Max number of documents in one request"""
        return 10



# TODO this should not be needed in the future
name_to_id_map = {
    'NOWPreprocessor': 'vlxgoxlv',
    'NOWQdrantIndexer16': '11yxatzh',
    'NOWElasticIndexer': 'mdtuyaaj',
    'NOWAutoCompleteExecutor2': 'w5w084h7',
}



""" Module contains data-transfer-object for finetune datasets. """
from dataclasses import dataclass
from typing import Dict, Optional

from docarray import Document, DocumentArray

from now.finetuning.settings import FinetuneSettings

_SEED = 42
MAX_VAL_SET_SIZE = 5000


@dataclass
class FinetuneDataset:
    index: DocumentArray

    train: Optional[DocumentArray] = None
    val: Optional[DocumentArray] = None
    val_query: Optional[DocumentArray] = None
    val_index: Optional[DocumentArray] = None

    def as_dict(self) -> Dict[str, Optional[DocumentArray]]:
        return {
            'train': self.train,
            'val': self.val,
            'val_query': self.val_query,
            'val_index': self.val_index,
        }


def build_finetuning_dataset(
    dataset: DocumentArray, finetune_settings: FinetuneSettings
) -> FinetuneDataset:
    ds = FinetuneDataset(index=dataset)

    split_index = max(
        int(len(dataset) * finetune_settings.train_val_split_ration),
        len(dataset) - MAX_VAL_SET_SIZE,
    )

    def _create_finetune_subset(subset: DocumentArray) -> DocumentArray:
        if finetune_settings.add_embeddings:
            return DocumentArray(
                [
                    Document(
                        tensor=doc.embedding.astype('float32'),
                        tags={'finetuner_label': doc.tags['finetuner_label']},
                    )
                    for doc in subset
                ]
            )
        else:
            return subset

    ds.train = _create_finetune_subset(dataset[:split_index])
    ds.val = _create_finetune_subset(dataset[split_index:])

    ds.val_index = _create_finetune_subset(dataset[split_index:])
    ds.val_query = _create_finetune_subset(
        dataset[split_index:].sample(k=finetune_settings.num_val_queries, seed=_SEED)
    )

    return ds



from collections import defaultdict

from docarray import DocumentArray


def merge_matches_sum(query_docs, limit):
    # in contrast to merge_matches_min, merge_matches_avg sorts the parent matches by the average distance of all chunk matches
    # we have 3 chunks indexed for each root document but the matches might contain less than 3 chunks
    # in case of less than 3 chunks, we assume that the distance of the missing chunks is the same to the last match
    # m.score.value is a distance metric
    query_doc = query_docs[0]
    parent_id_count_and_sum_and_chunks = defaultdict(lambda: [0, 0, []])
    for m in query_doc.matches:
        count_and_sum_and_chunks = parent_id_count_and_sum_and_chunks[m.parent_id]
        distance = m.scores['cosine'].value
        count_and_sum_and_chunks[0] += 1
        count_and_sum_and_chunks[1] += distance
        count_and_sum_and_chunks[2].append(m)
    all_matches = DocumentArray()
    for group in (3, 2, 1):
        parent_id_to_sum_and_chunks = {
            parent_id: count_and_sum_and_chunks[1:]
            for parent_id, count_and_sum_and_chunks in parent_id_count_and_sum_and_chunks.items()
            if count_and_sum_and_chunks[0] == group
        }
        parent_to_sum_sorted = sorted(
            parent_id_to_sum_and_chunks.items(), key=lambda x: x[1][0]
        )
        matches = [sum_and_chunks[1][0] for _, sum_and_chunks in parent_to_sum_sorted]
        all_matches.extend(matches)
        print(f'# num parents for group {group}: {len(matches)}')
    query_doc.matches = all_matches[:limit]



import itertools
import math
import os
from inspect import getmembers, isclass
from typing import List, Optional, Tuple

from dask.distributed import Client
from docarray import Document, DocumentArray

from now.finetuning import generation_fns
from now.finetuning.generation_fns import GeneratorFunction
from now.now_dataclasses import Task, TrainDataGenerationConfig


class EncoderDataBuilder:
    def __init__(
        self,
        name: str,
        methods: List[TrainDataGenerationConfig],
        encoder_type: str,
    ):
        """
        Fine-tuning data generation for a specific encoder.

        :param name: Name of the encoder.
        :param methods: List of methods to generate fine-tuning data.
        :param encoder_type: Type of the encoder, either `single_model`
            or `multi_model`.
        """
        self._name = name
        self._methods = methods
        self._encoder_type = encoder_type
        self._modality = (
            'multi_model' if self._encoder_type == 'text-to-image' else 'single_model'
        )
        self._generation_cls = {
            cls.name(): cls
            for _, cls in getmembers(generation_fns, isclass)
            if issubclass(cls, GeneratorFunction)
        }

    def build(self, es_data: DocumentArray) -> DocumentArray:
        """
        Generates query and target data based on method(s).

        In case the encoder is `single_model`, we save generated targets and queries
        separately. If the encoder is `multi_model`, we generate query-target pairs.

        :param es_data: Data extracted from ES.
        :return: `DocumentArray` of generated data.
        """
        data = DocumentArray()
        for method in self._methods:
            query_generator = self._generation_cls[method.query.method](
                scope=method.query.scope, **method.query.parameters
            )
            target_generator = self._generation_cls[method.target.method](
                scope=method.target.scope, **method.target.parameters
            )
            for document in es_data:
                queries = query_generator.process(document=document)
                targets = target_generator.process(document=document)
                if self._modality == 'single_model':
                    merged_docs = []
                    for doc in [*queries, *targets]:
                        doc.tags = {'finetuner_label': document.id}
                        merged_docs.append(doc)
                    data.extend(merged_docs)
                else:
                    data.extend(
                        [
                            Document(chunks=[query, target])
                            for query, target in itertools.product(queries, targets)
                        ]
                    )
        return data

    @property
    def name(self):
        return self._name

    @property
    def encoder_type(self):
        return self._encoder_type


class DataBuilder:
    def __init__(
        self,
        dataset: DocumentArray,
        config: Task,
        num_workers: Optional[int] = None,
        threads_per_worker: Optional[int] = 4,
    ):
        """
        Fine-tuning data generation for a specific task.

        :param dataset: A `DocumentArray` extracted from ES.
        :param config: A `TaskConfig` object containing information about
            data generation methods.
        :param num_workers: Number of workers for data generation.
        :param threads_per_worker: Number of threads per worker.
        """
        self._dataset = dataset
        self._num_workers = num_workers if num_workers else os.cpu_count()
        self._enc_data_builders = self._init_enc_data_builders(config)
        self._dask_client = Client(
            threads_per_worker=threads_per_worker,
            n_workers=self._num_workers,
        )

    @staticmethod
    def _init_enc_data_builders(
        config: Task,
    ) -> List[Tuple[str, EncoderDataBuilder]]:
        """
        Initialize data builders for encoder(s).

        These builders will be used when `DataGenerator.build()` is called.

        :param config: A `TaskConfig` object containing information about ES data and
            data generation approaches.
        :return: List of pairs - intended dataset name and `EncoderDataBuilder`.
        """
        return [
            (
                encoder.train_dataset_name,
                EncoderDataBuilder(
                    name=encoder.name,
                    methods=encoder.training_data_generation_methods,
                    encoder_type=encoder.encoder_type,
                ),
            )
            for encoder in config.encoders
        ]

    @staticmethod
    def _data_chunk_generator(data: DocumentArray, num_chunks: int) -> DocumentArray:
        """Generates chunks of ES data for parallel data generation.

        :param data: ES data.
        :param num_chunks: number of chunks to be generated.
        :return: Data chunk as a `DocumentArray`.
        """
        chunk_size = math.ceil(len(data) / num_chunks)
        for pos in range(0, len(data), chunk_size):
            yield data[pos : pos + chunk_size]  # noqa: E203

    def build(
        self,
        to_hubble: bool = False,
        data_dir: Optional[str] = None,
    ) -> List[Tuple[DocumentArray, str]]:
        """
        Generates data from ES dataset based on the task configuration file.

        You can also upload the generated data on hubble, or save it locally
        (or both).

        :param to_hubble: Uploads data to Hubble if `True`.
        :param data_dir: Saves data locally in the given directory if it's not `None`.
        :return: Generated data for each encoder.
        """

        def _generate(data: DocumentArray, data_builder: EncoderDataBuilder):
            return data_builder.build(es_data=data)

        data = []
        for dataset_name, enc_data_builder in self._enc_data_builders:
            futures = []
            for chunk in self._data_chunk_generator(
                data=self._dataset, num_chunks=self._num_workers
            ):
                futures.append(
                    self._dask_client.submit(_generate, chunk, enc_data_builder)
                )
            results = self._dask_client.gather(futures)
            dataset = DocumentArray(itertools.chain.from_iterable(results))

            data.append((dataset, enc_data_builder.encoder_type))

            if to_hubble:
                dataset.push(dataset_name, show_progress=True, public=False)
            if data_dir:
                os.makedirs(data_dir, exist_ok=True)
                dataset.save_binary(os.path.join(data_dir, dataset_name))
        return data



import os

import cowsay
import requests
from rich import box
from rich.console import Console
from rich.panel import Panel
from rich.table import Column, Table

from now import run_backend, run_bff_playground
from now.cloud_manager import setup_cluster
from now.constants import DOCKER_BFF_PLAYGROUND_TAG, SURVEY_LINK, DatasetTypes
from now.deployment.deployment import cmd, list_all_wolf, status_wolf, terminate_wolf
from now.dialog import configure_user_input
from now.log import yaspin_extended
from now.system_information import get_system_state
from now.utils import _get_context_names, get_flow_id, maybe_prompt_user, sigmap


def stop_now(contexts, active_context, **kwargs):
    choices = _get_context_names(contexts, active_context)
    # Add all remote Flows that exists with the namespace `nowapi`
    alive_flows = list_all_wolf()
    for flow_details in alive_flows:
        choices.append(flow_details['gateway'])
    if len(choices) == 0:
        cowsay.cow('nothing to stop')
        return
    else:
        questions = [
            {
                'type': 'list',
                'name': 'cluster',
                'message': 'Which cluster do you want to delete?',
                'choices': choices,
            }
        ]
        cluster = maybe_prompt_user(questions, 'cluster', **kwargs)
    if cluster == 'kind-jina-now':
        delete_cluster = maybe_prompt_user(
            [
                {
                    'type': 'list',
                    'name': 'delete-cluster',
                    'message': 'Do you want to delete the entire cluster or just the namespace?',
                    'choices': [
                        {'name': '⛔ no, keep the cluster', 'value': False},
                        {'name': '✅ yes, delete everything', 'value': True},
                    ],
                }
            ],
            attribute='delete-cluster',
            **kwargs,
        )
        if delete_cluster:
            with yaspin_extended(
                sigmap=sigmap, text=f"Remove local cluster {cluster}", color="green"
            ) as spinner:
                cmd(f'{kwargs["kind_path"]} delete clusters jina-now')
                spinner.ok('💀')
            cowsay.cow('local jina NOW cluster removed')
        else:
            with yaspin_extended(
                sigmap=sigmap,
                text=f"Remove namespace nowapi NOW from {cluster}",
                color="green",
            ) as spinner:
                cmd(f'{kwargs["kubectl_path"]} delete ns nowapi')
                spinner.ok('💀')
            cowsay.cow(f'nowapi namespace removed from {cluster}')
    elif 'wolf.jina.ai' in cluster:
        flow = [x for x in alive_flows if x['gateway'] == cluster][0]
        flow_id = get_flow_id(flow['name'])
        _result = status_wolf(flow_id)
        if _result is None:
            print(f'❎ Flow not found in JCloud. Likely, it has been deleted already')
        if _result is not None and _result['status'] == 'ALIVE':
            terminate_wolf(flow_id)
            from hubble import Client

            cookies = {'st': Client().token}
            requests.delete(
                f'https://storefrontapi.nowrun.jina.ai/api/v1/schedule_sync/{flow_id}',
                cookies=cookies,
            )
        cowsay.cow(f'remote Flow `{cluster}` removed')
    else:
        with yaspin_extended(
            sigmap=sigmap, text=f"Remove jina NOW from {cluster}", color="green"
        ) as spinner:
            cmd(f'{kwargs["kubectl_path"]} delete ns nowapi')
            spinner.ok('💀')
        cowsay.cow(f'nowapi namespace removed from {cluster}')


def get_task(kwargs):
    for x in ['cli', 'now']:
        if x in kwargs:
            return kwargs[x]
    raise Exception('kwargs do not contain a task')


def start_now(**kwargs):
    user_input = configure_user_input(**kwargs)
    app_instance = user_input.app_instance
    # Only if the deployment is remote and the demo examples is available for the selected app
    # Should not be triggered for CI tests
    if app_instance.is_demo_available(user_input):
        gateway_host = 'remote'
        gateway_host_internal = f'grpcs://now-example-{app_instance.app_name}-{user_input.dataset_name}.dev.jina.ai'.replace(
            '_', '-'
        )
        gateway_port_internal = None
    else:
        if not os.environ.get('NOW_TESTING', False):
            setup_cluster(user_input, **kwargs)
        (
            gateway_host,
            gateway_port,
            gateway_host_internal,
            gateway_port_internal,
        ) = run_backend.run(app_instance, user_input, **kwargs)

    if os.environ.get('NOW_TESTING', False):
        # start_bff(9090, daemon=True)
        # sleep(10)
        bff_playground_host = 'http://localhost'
        bff_port = '9090'
        playground_port = '80'
    elif gateway_host == 'localhost' or 'NOW_CI_RUN' in os.environ:
        # only deploy playground when running locally or when testing
        bff_playground_host, bff_port, playground_port = run_bff_playground.run(
            gateway_host=gateway_host,
            docker_bff_playground_tag=DOCKER_BFF_PLAYGROUND_TAG,
            kubectl_path=kwargs['kubectl_path'],
        )
    else:
        bff_playground_host = 'https://nowrun.jina.ai'
        bff_port = '80'
        playground_port = '80'
    # TODO: add separate BFF endpoints in print output
    input_modality = '-or-'.join(app_instance.input_modality)
    output_modality = '-or-'.join(app_instance.output_modality)
    bff_url = (
        bff_playground_host
        + ('' if str(bff_port) == '80' else f':{bff_port}')
        + f'/api/v1/{input_modality}-to-{output_modality}/docs'
    )
    playground_url = (
        bff_playground_host
        + ('' if str(playground_port) == '80' else f':{playground_port}')
        + (
            f'/?host='
            + (gateway_host_internal if gateway_host != 'localhost' else 'gateway')
            + f'&input_modality={input_modality}'
            + f'&output_modality={output_modality}'
            + (
                f'&data={user_input.dataset_name if user_input.dataset_type == DatasetTypes.DEMO else "custom"}'
            )
            + (f'&secured={user_input.secured}' if user_input.secured else '')
        )
        + (f'&port={gateway_port_internal}' if gateway_port_internal else '')
    )
    print()
    my_table = Table(
        'Attribute',
        Column(header="Value", overflow="fold"),
        show_header=False,
        box=box.SIMPLE,
        highlight=True,
    )
    my_table.add_row('Api docs', bff_url)
    if user_input.secured and user_input.api_key:
        my_table.add_row('API Key', user_input.api_key)
    my_table.add_row('Playground', playground_url)
    console = Console()
    console.print(
        Panel(
            my_table,
            title=f':tada: {input_modality}-to-{output_modality} app is NOW ready!',
            expand=False,
        )
    )
    return {
        'bff': bff_url,
        'playground': playground_url,
        'bff_playground_host': bff_playground_host,
        'bff_port': bff_port,
        'playground_port': playground_port,
        'input_modality': input_modality,
        'output_modality': output_modality,
        'host': gateway_host_internal,
        'port': gateway_port_internal,
        'secured': user_input.secured,
    }


def run_k8s(os_type: str = 'linux', arch: str = 'x86_64', **kwargs):
    contexts, active_context = get_system_state(**kwargs)
    task = get_task(kwargs)
    if task == 'start':
        return start_now(
            contexts=contexts,
            active_context=active_context,
            os_type=os_type,
            arch=arch,
            **kwargs,
        )
    elif task == 'stop':
        return stop_now(contexts, active_context, **kwargs)
    elif task == 'survey':
        import webbrowser

        webbrowser.open(SURVEY_LINK, new=0, autoraise=True)
    else:
        raise Exception(f'unknown task, {task}')


if __name__ == '__main__':
    run_k8s(
        app='text_to_image',
        data='bird-species',
        cluster='kind-jina-now',
        deployment_type='local',
        kubectl_path='/usr/local/bin/kubectl',
        kind_path='/Users/sebastianlettner/.cache/jina-now/kind',
        proceed=True,
        now='start',
    )



import traceback
from typing import Any, Dict, List, Mapping, Optional, Union

import numpy as np
from docarray import Document, DocumentArray
from docarray.score import NamedScore
from elasticsearch import Elasticsearch
from elasticsearch.helpers import bulk

from now.constants import ModelDimensions
from now.executor.abstract.auth import (
    SecurityLevel,
    get_auth_executor_class,
    secure_request,
)

metrics_mapping = {
    'cosine': 'cosineSimilarity',
    'l2_norm': 'l2norm',
}

Executor = get_auth_executor_class()


class NOWElasticIndexer(Executor):
    def __init__(
        self,
        dims: Union[List[int], int],
        hosts: Union[
            str, List[Union[str, Mapping[str, Union[str, int]]]], None
        ] = 'https://elastic:elastic@localhost:9200',
        es_config: Optional[Dict[str, Any]] = None,
        metric: str = 'cosine',
        index_name: str = 'now-index',
        **kwargs,
    ):
        """
        Initializer function for the NOWElasticIndexer.

        :param hosts: host configuration of the Elasticsearch node or cluster
        :param es_config: Elasticsearch cluster configuration object
        :param metric: The distance metric used for the vector index and vector search
        :param dims: The dimensions of your embeddings.
        :param index_name: ElasticSearch Index name used for the storage
        :param es_mapping: Mapping for new index. If none is specified, this will be
            generated from metric and dims. Embeddings from chunk documents will
            always be stored in fields `embedding_x` where x iterates over the number
            of embedding fields (length of `dims`) to be created in the index.
                (used for indexing, delete and update), e.g. '@r', '@c', '@r,c'.
        """
        super().__init__(**kwargs)

        self.hosts = hosts
        self.metric = metric
        self.index_name = index_name
        self.es_config = {'verify_certs': False} if not es_config else es_config
        self.dims = dims if isinstance(dims, list) else [dims]
        self.es_mapping = self._generate_es_mapping(dims)
        self.es = Elasticsearch(hosts=self.hosts, **self.es_config, ssl_show_warn=False)
        if not self.es.indices.exists(index=self.index_name):
            self.es.indices.create(index=self.index_name, mappings=self.es_mapping)

    def _generate_es_mapping(self, dims: List[int]) -> Dict:
        es_mapping = {
            'properties': {
                'id': {'type': 'keyword'},
                'bm25_text': {'type': 'text', 'analyzer': 'standard'},
            }
        }

        for i, dim in enumerate(self.dims):
            es_mapping['properties'][f'chunk_{i}'] = {
                'properties': {
                    f'embedding': {
                        'type': 'dense_vector',
                        'dims': dim,
                        'similarity': self.metric,
                        'index': 'true',
                    }
                }
            }
        return es_mapping

    @secure_request(on='/index', level=SecurityLevel.USER)
    def index(
        self,
        docs: DocumentArray,
        docs_matrix: List[DocumentArray] = None,
        parameters: dict = None,
        **kwargs,
    ) -> DocumentArray:
        """
        Index new `Document`s by adding them to the Elasticsearch index.

        :param docs: Documents to be indexed.
        :param parameters: dictionary with options for indexing.
        :return: empty `DocumentArray`.
        """
        if not docs:
            return docs
        if docs_matrix:
            if len(docs_matrix) > 1:
                docs = self._join_docs_matrix_into_chunks(
                    docs_matrix=docs_matrix, on='index'
                )
            else:
                docs = docs_matrix[0]
        if not parameters:
            parameters = {}
        es_docs = self._transform_docs_to_es(docs)
        try:
            success, _ = bulk(self.es, es_docs)
            self.es.indices.refresh(index=self.index_name)
        except Exception as e:
            print(e)
            print(traceback.format_exc())
            raise
        if success:
            print(
                f'Inserted {success} documents into Elasticsearch index {self.index_name}'
            )
        return (
            DocumentArray()
        )  # prevent sending the data back by returning an empty DocumentArray

    @secure_request(on='/search', level=SecurityLevel.USER)
    def search(
        self,
        docs: Union[Document, DocumentArray],
        docs_matrix: Optional[List[DocumentArray]] = None,
        parameters: dict = {},
        **kwargs,
    ):
        """Perform traditional bm25 + vector search. By convention, BM25 will search on
        the 'bm25_text' field of the index. For now, this field contains a concatenation of
        all text chunks of the documents.

        Search can be performed with candidate filtering. Filters are a triplet (column,operator,value).
        More than a filter can be applied during search. Therefore, conditions for a filter are specified as a list triplets.
        Each triplet contains:
            - field: Field used to filter.
            - operator: Binary operation between two values. Some supported operators include `['>','<','=','<=','>=']`.
            - value: value used to compare a candidate.

        :param docs: query `Document`s.
        :param parameters: dictionary of options for searching.
            Keys accepted:
                - 'filter' (dict): the filtering conditions on document tags
                - 'limit' (int): nr of matches to get per Document
        """
        if not docs:
            return docs
        if docs_matrix:
            if len(docs_matrix) > 1:
                docs = self._join_docs_matrix_into_chunks(
                    docs_matrix=docs_matrix, on='search'
                )
            else:
                docs = docs_matrix[0]
        search_filter = parameters.get('filter', None)
        limit = parameters.get('limit', 20)
        apply_bm25 = parameters.get('apply_bm25', False)
        for doc in docs:
            query = self._build_es_query(
                doc=doc,
                apply_bm25=apply_bm25,
                search_filter=search_filter,
            )
            try:
                result = self.es.search(
                    index=self.index_name,
                    query=query,
                    source=True,
                    size=limit,
                )['hits']['hits']
                doc.matches = self._transform_es_results_to_matches(result)
            except Exception:
                print(traceback.format_exc())
        return docs

    @secure_request(on='/update', level=SecurityLevel.USER)
    def update(self, docs: DocumentArray, **kwargs) -> DocumentArray:
        """
        TODO: implement update endpoint, eg. update ES docs with new embeddings etc.
        """
        raise NotImplementedError()

    @secure_request(on='/list', level=SecurityLevel.USER)
    def list(self, parameters: dict = {}, **kwargs):
        """List all indexed documents.

        Note: this implementation is naive and does not
        consider the default maximum documents in a page returned by `Elasticsearch`.
        Should be addressed in future with `scroll`.

        :param parameters: dictionary with limit and offset
        - offset (int): number of documents to skip
        - limit (int): number of retrieved documents
        """
        limit = int(parameters.get('limit', 20))
        offset = int(parameters.get('offset', 0))
        try:
            result = self.es.search(index=self.index_name, query={'match_all': {}})[
                'hits'
            ]['hits']
        except Exception:
            print(traceback.format_exc())
        if result:
            result_da = self._transform_es_to_da(result)
            return result_da[offset : offset + limit]
        else:
            return result

    @secure_request(on='/delete', level=SecurityLevel.USER)
    def delete(self, parameters: dict = {}, **kwargs):
        """
        Delete endpoint to delete document/documents from the index.

        :param parameters: dictionary with filter conditions to select
            documents for deletion.
        """
        search_filter = parameters.get('filter', None)
        if search_filter:
            es_search_filter = {'query': {'bool': {}}}
            for field, filters in search_filter.items():
                for operator, filter in filters.items():
                    if isinstance(filter, str):
                        es_search_filter['query']['bool']['filter'] = {
                            'term': {'tags.' + field: filter}
                        }
                    elif isinstance(filter, int) or isinstance(filter, float):
                        operator = operator.replace('$', '')
                        es_search_filter['query']['bool']['filter'] = {
                            'range': {'tags.' + field: {operator: filter}}
                        }
        try:
            resp = self.es.delete_by_query(index=self.index_name, body=es_search_filter)
            self.es.indices.refresh(index=self.index_name)
        except Exception:
            print(traceback.format_exc())
            raise
        if resp:
            print(
                f"Deleted {resp['deleted']} documents in Elasticsearch index {self.index_name}"
            )
        return DocumentArray()

    def _build_es_query(
        self,
        doc: Document,
        apply_bm25: bool,
        search_filter: Optional[Dict] = None,
    ) -> Dict:
        """
        Build script-score query used in Elasticsearch. To do this, we extract
        embeddings from the query document and pass them in the s   cript-score
        query together with the fields to search on in the Elasticsearch index.

        :param query: a `Document` with chunks containing a text embedding and
            image embedding.
        :param apply_bm25: whether to combine bm25 with vector search. If False,
            will only perform vector search queries. If True, must supply a text
            field for bm25 searching.
        :param search_filter: dictionary of filters to apply to the search.
        :return: a dictionary containing query and filter.
        """
        source = ''
        query = {
            'bool': {
                'should': [
                    {'match_all': {}},
                ],
            },
        }

        # build bm25 part
        if apply_bm25:
            source += '_score / (_score + 10.0) + '
            text = doc.text
            multi_match = {'multi_match': {'query': text, 'fields': ['bm25_text']}}
            query['bool']['should'].append(multi_match)

        # add filter
        if search_filter:
            es_search_filter = {}
            for field, filters in search_filter.items():
                for operator, filter in filters.items():
                    if isinstance(filter, str):
                        es_search_filter['term'] = {"tags." + field: filter}
                    elif isinstance(filter, int) or isinstance(filter, float):
                        operator = operator.replace('$', '')
                        es_search_filter['range'] = {
                            "tags." + field: {operator: filter}
                        }
            query['bool']['filter'] = es_search_filter

        # build vector search part
        query_embeddings = self._extract_embeddings(doc)
        params = {}
        for key, embedding in query_embeddings.items():
            if key == 'embedding':
                source += f"0.5*{metrics_mapping[self.metric]}(params.query_{key}, '{key}') + "
            else:
                source += f"0.5*{metrics_mapping[self.metric]}(params.query_{key}, '{key}.embedding') + "
            params[f'query_{key}'] = embedding
        source += '1.0'
        query_json = {
            'script_score': {
                'query': query,
                'script': {'source': source, 'params': params},
            }
        }
        return query_json

    def _transform_es_to_da(self, result: Union[Dict, List[Dict]]) -> DocumentArray:
        """
        Transform Elasticsearch documents into DocumentArray. Assumes that all Elasticsearch
        documents have a 'text' field. It does not return embeddings as part of the Document.

        :param result: results from an Elasticsearch query.
        :return: a DocumentArray containing all results.
        """
        if isinstance(result, Dict):
            result = [result]
        da = DocumentArray()
        for es_doc in result:
            doc = Document(id=es_doc['_id'])
            for k, v in es_doc['_source'].items():
                if k.startswith('chunk'):
                    chunk = Document.from_dict(v)
                    doc.chunks.append(chunk)
                elif k.startswith('embedding'):
                    continue
                elif k in ['bm25_text', '_score']:
                    continue
                else:
                    doc.k = v
            da.append(doc)
        return da

    def _transform_docs_to_es(self, docs: DocumentArray) -> List[Dict]:
        """
        This function takes Documents as input and transforms them into a list of
        dictionaries that can be indexed in Elasticsearch.

        :param docs: documents containing text and image chunks.
        :return: list of dictionaries containing text, text embedding and image embedding
        """
        es_docs = list()
        for doc in docs:
            es_doc = {k: v for k, v in doc.to_dict().items() if v}
            es_doc['_id'] = doc.id
            es_doc['bm25_text'] = doc.text
            chunks = es_doc.pop('chunks', None)
            for i, chunk in enumerate(chunks):
                es_doc[f'chunk_{i}'] = {k: v for k, v in chunk.items() if v}
                if chunk['text']:
                    es_doc['bm25_text'] += " " + chunk['text']
            es_doc['_op_type'] = 'index'
            es_doc['_index'] = self.index_name
            es_docs.append(es_doc)
        return es_docs

    def _transform_es_results_to_matches(self, es_results: List[Dict]) -> DocumentArray:
        """
        Transform a list of results from Elasticsearch into a matches in the form of a `DocumentArray`.

        :param es_results: List of dictionaries containing results from Elasticsearch querying.
        :return: `DocumentArray` that holds all matches in the form of `Document`s.
        """
        matches = DocumentArray()
        for result in es_results:
            d = self._transform_es_to_da(result)[0]
            d.scores[self.metric] = NamedScore(value=result['_score'])
            matches.append(d)
        return matches

    @staticmethod
    def _extract_embeddings(doc: Document) -> Dict[str, np.ndarray]:
        """
        Get embeddings from a documents.

        :param doc: `Document` with chunks of text document and/or image document.
        :return: Embeddings as values in a dictionary, modality specified in key.
        """
        embeddings = {}

        for i, chunk in enumerate(doc.chunks):
            embeddings[f"chunk_{i}"] = chunk.embedding
        if not embeddings:
            print('No embeddings extracted')
            raise
        return embeddings

    @staticmethod
    def _join_docs_matrix_into_chunks(
        docs_matrix: List[DocumentArray], on: str = 'index'
    ) -> DocumentArray:
        """
        Transform a matrix of DocumentArray's into one DocumentArray, by adding Documents to the chunk level.
        If we are joining a docs_matrix during searching, we will keep both the clip text embedding
        and sbert embedding generated for the text document chunk. Otherwise, if we are joining a docs_matrix
        during indexing, we will only keep the sbert embedding and the image embedding generated by clip as
        chunk documents.

        :param on: str = 'index',
        :param on: the endpoint which is called, if 'index', then we keep sbert embedding
            and clip image embedding; if 'search', we keep sbert embedding and clip text embedding.
        """
        new_da = DocumentArray()
        if on == 'search':
            for doc1, doc2 in zip(*docs_matrix):
                new_doc = Document(text=doc1.chunks[0].text)
                if (
                    len(doc1.chunks[0].embedding) == ModelDimensions.SBERT
                    and len(doc2.chunks[0].embedding) == ModelDimensions.CLIP
                ):
                    new_doc.chunks.extend(doc1.chunks)
                    new_doc.chunks.extend(doc2.chunks)
                elif (
                    len(doc1.chunks[0].embedding) == ModelDimensions.CLIP
                    and len(doc2.chunks[0].embedding) == ModelDimensions.SBERT
                ):
                    new_doc.chunks.extend(doc2.chunks)
                    new_doc.chunks.extend(doc1.chunks)
                else:
                    raise Exception('Embedding size not supported')
                new_da.append(new_doc)
        else:
            for doc1, doc2 in zip(*docs_matrix):
                new_doc = Document()
                text_chunks = [
                    c for c in doc1.chunks if c.text
                ]  # doc1 from SBert with text embedding
                image_chunks = [
                    c for c in doc2.chunks if c.uri
                ]  # doc2 from CLIP with image embedding
                new_doc.chunks.extend(text_chunks)
                new_doc.chunks.extend(image_chunks)
                new_da.append(new_doc)

        return new_da



from typing import List, Optional

from pydantic import Field

from deployment.bff.app.v1.models.helper import (
    BaseIndexRequestModel,
    BaseSearchRequestModel,
    BaseSearchResponseModel,
)


# Request Model
class NowTextImageIndexRequestModel(BaseIndexRequestModel):
    texts: Optional[List[str]] = Field(
        default=[], description='List of Texts to index.'
    )
    images: Optional[List[str]] = Field(
        default=[],
        description='Image query. Image should be base64encoded in `utf-8` format',
    )


class NowTextImageSearchRequestModel(BaseSearchRequestModel):
    image: Optional[str] = Field(
        default=None,
        description='Image query. Image should be base64encoded in `utf-8` format',
    )
    text: Optional[str] = Field(default=None, description='Text query')


# Response Model
class NowTextImageResponseModel(BaseSearchResponseModel):
    text: Optional[str] = Field(description='Matching text result.', default='')
    blob: Optional[str] = Field(
        description='Base64 encoded image in `utf-8` str format'
    )


NowTextImageResponseModel.update_forward_refs()



import base64
from itertools import zip_longest
from typing import List

from docarray import Document, DocumentArray
from fastapi import APIRouter

from deployment.bff.app.v1.models.text import NowTextSearchRequestModel
from deployment.bff.app.v1.models.video import (
    NowVideoIndexRequestModel,
    NowVideoResponseModel,
)
from deployment.bff.app.v1.routers.helper import jina_client_post, process_query

router = APIRouter()


# Index
@router.post(
    "/index",
    summary='Add more video data to the indexer',
)
def index(data: NowVideoIndexRequestModel):
    """
    Append the list of video data to the indexer. Each video data should be
    `base64` encoded using human-readable characters - `utf-8`.
    """
    index_docs = DocumentArray()
    for video, uri, tags in zip_longest(data.videos, data.uris, data.tags):
        if bool(video) + bool(uri) != 1:
            raise ValueError(
                f'Can only set one value but have video={video}, uri={uri}'
            )
        if video:
            base64_bytes = video.encode('utf-8')
            message = base64.decodebytes(base64_bytes)
            index_docs.append(Document(blob=message, tags=tags, modality='video'))
        else:
            index_docs.append(Document(uri=uri, tags=tags, modality='video'))

    jina_client_post(
        data=data,
        inputs=index_docs,
        endpoint='/index',
    )


# Search
@router.post(
    "/search",
    response_model=List[NowVideoResponseModel],
    summary='Search video data via text as query',
)
def search(data: NowTextSearchRequestModel):
    """
    Retrieve matching videos for a given text as query.
    """
    query_doc, filter_query = process_query(
        text=data.text, uri=data.uri, conditions=data.filters
    )

    docs = jina_client_post(
        data=data,
        inputs=query_doc,
        parameters={
            'limit': data.limit,
            'filter': filter_query,
        },
        endpoint='/search',
    )
    return docs[0].matches[: data.limit].to_dict()


@router.post(
    "/suggestion",
    summary='Get auto complete suggestion for query',
)
def suggestion(data: NowTextSearchRequestModel):
    """
    Return text suggestions for the rest of the query text
    """
    query_doc, _ = process_query(text=data.text, uri=data.uri, conditions=data.filters)
    docs = jina_client_post(
        data=data,
        inputs=query_doc,
        endpoint='/suggestion',
    )
    return docs.to_dict()



from __future__ import annotations, print_function, unicode_literals

from now.utils import BetterEnum

# TODO: Uncomment the DEMO_DATASET_DOCARRAY_VERSION when the DocumentArray datasets on GCloud has been changed
# from docarray import __version__ as docarray_version

DEMO_DATASET_DOCARRAY_VERSION = '0.13.17'

DOCKER_BFF_PLAYGROUND_TAG = '0.0.135-test-reactivate'
NOW_PREPROCESSOR_VERSION = '0.0.108-test-reactivate'
NOW_QDRANT_INDEXER_VERSION = '0.0.9-refactor-improve-performance-curation-3'
NOW_ELASTIC_INDEXER_VERSION = '0.0.8-refactor-improve-performance-curation-3'
NOW_AUTOCOMPLETE_VERSION = '0.0.4-refactor-simplify-preprocessing-45'


class Modalities(BetterEnum):
    TEXT = 'text'
    IMAGE = 'image'
    MUSIC = 'music'
    VIDEO = 'video'
    TEXT_AND_IMAGE = 'text-and-image'


class Apps(BetterEnum):
    IMAGE_TEXT_RETRIEVAL = 'image_text_retrieval'
    MUSIC_TO_MUSIC = 'music_to_music'
    TEXT_TO_VIDEO = 'text_to_video'
    TEXT_TO_TEXT_AND_IMAGE = 'text_to_text_and_image'


class DatasetTypes(BetterEnum):
    DEMO = 'demo'
    PATH = 'path'
    URL = 'url'
    DOCARRAY = 'docarray'
    S3_BUCKET = 's3_bucket'
    ELASTICSEARCH = 'elasticsearch'


class ModelNames(BetterEnum):
    MLP = 'mlp'
    SBERT = 'sentence-transformers/msmarco-distilbert-base-v3'
    CLIP = 'openai/clip-vit-base-patch32'


class ModelDimensions(BetterEnum):
    SBERT = 768
    CLIP = 512


SUPPORTED_FILE_TYPES = {
    Modalities.TEXT: ['txt', 'md'],
    Modalities.IMAGE: ['jpg', 'jpeg', 'png', 'gif', 'bmp', 'tiff', 'tif'],
    Modalities.MUSIC: ['mp3', 'wav', 'ogg', 'flac'],
    Modalities.VIDEO: ['gif'],
}

BASE_STORAGE_URL = (
    'https://storage.googleapis.com/jina-fashion-data/data/one-line/datasets'
)

CLIP_USES = {
    'local': ('CLIPOnnxEncoder/latest', 'ViT-B-32::openai', ModelDimensions.CLIP),
    'remote': ('CLIPOnnxEncoder/latest-gpu', 'ViT-B-32::openai', ModelDimensions.CLIP),
}
EXTERNAL_CLIP_HOST = 'encoderclip-climbing-lobster-885354f905.wolf.jina.ai'

DEFAULT_FLOW_NAME = 'nowapi'
PREFETCH_NR = 10

SURVEY_LINK = 'https://10sw1tcpld4.typeform.com/to/VTAyYRpR?utm_source=cli'

TAG_OCR_DETECTOR_TEXT_IN_DOC = '_ocr_detector_text_in_doc'
TAG_INDEXER_DOC_HAS_TEXT = '_indexer_doc_has_text'
EXECUTOR_PREFIX = 'jinahub+docker://'
ACCESS_PATHS = '@cc'



import json
import os
import pathlib
import tempfile
import time
from copy import deepcopy
from os.path import expanduser as user
from typing import Dict, List, Optional, Tuple

import hubble
from docarray import Document, DocumentArray
from jina import __version__ as jina_version
from jina.helper import random_port

from now.app.base.app import JinaNOWApp
from now.constants import (
    EXECUTOR_PREFIX,
    EXTERNAL_CLIP_HOST,
    NOW_AUTOCOMPLETE_VERSION,
    NOW_ELASTIC_INDEXER_VERSION,
    NOW_PREPROCESSOR_VERSION,
    NOW_QDRANT_INDEXER_VERSION,
    PREFETCH_NR,
    TAG_INDEXER_DOC_HAS_TEXT,
    DatasetTypes,
    Modalities,
)
from now.demo_data import DEFAULT_EXAMPLE_HOSTED
from now.deployment.deployment import cmd
from now.executor.name_to_id_map import name_to_id_map
from now.finetuning.run_finetuning import finetune
from now.finetuning.settings import FinetuneSettings, parse_finetune_settings
from now.now_dataclasses import UserInput
from now.utils import maybe_download_from_s3

cur_dir = pathlib.Path(__file__).parent.resolve()


MAX_RETRIES = 20


def common_get_flow_env_dict(
    finetune_settings: FinetuneSettings,
    encoder_uses: str,
    encoder_with: Dict,
    encoder_uses_with: Dict,
    pre_trained_embedding_size: int,
    indexer_uses: str,
    indexer_resources: Dict,
    user_input: UserInput,
    tags: List,
):
    """Returns dictionary for the environments variables for the clip & music flow.yml files."""
    if (
        finetune_settings.perform_finetuning and finetune_settings.bi_modal
    ) or user_input.app_instance.app_name == 'music_to_music':
        pre_trained_embedding_size = pre_trained_embedding_size * 2

    config = {
        'JINA_VERSION': jina_version,
        'ENCODER_NAME': f'{EXECUTOR_PREFIX}{encoder_uses}',
        'CAST_CONVERT_NAME': f'{EXECUTOR_PREFIX}CastNMoveNowExecutor/v0.0.3',
        'N_DIM': finetune_settings.finetune_layer_size
        if finetune_settings.perform_finetuning
        or user_input.app_instance.app_name == 'music_to_music'
        else pre_trained_embedding_size,
        'PRE_TRAINED_EMBEDDINGS_SIZE': pre_trained_embedding_size,
        'INDEXER_NAME': f'{EXECUTOR_PREFIX}{indexer_uses}',
        'PREFETCH': PREFETCH_NR,
        'PREPROCESSOR_NAME': f'{EXECUTOR_PREFIX}{name_to_id_map.get("NOWPreprocessor")}/{NOW_PREPROCESSOR_VERSION}',
        'AUTOCOMPLETE_EXECUTOR_NAME': f'{EXECUTOR_PREFIX}{name_to_id_map.get("NOWAutoCompleteExecutor2")}/{NOW_AUTOCOMPLETE_VERSION}',
        'APP': user_input.app_instance.app_name,
        'COLUMNS': tags,
        'ADMIN_EMAILS': user_input.admin_emails or [] if user_input.secured else [],
        'USER_EMAILS': user_input.user_emails or [] if user_input.secured else [],
        'API_KEY': [user_input.api_key]
        if user_input.secured and user_input.api_key
        else [],
        **encoder_with,
        **indexer_resources,
    }

    if encoder_uses_with.get('pretrained_model_name_or_path'):
        config['PRE_TRAINED_MODEL_NAME'] = encoder_uses_with[
            "pretrained_model_name_or_path"
        ]
    if finetune_settings.perform_finetuning:
        config['FINETUNE_ARTIFACT'] = finetune_settings.finetuned_model_artifact
        config['JINA_TOKEN'] = finetune_settings.token

    config['CUSTOM_DNS'] = ''
    if 'NOW_EXAMPLES' in os.environ:
        valid_app = DEFAULT_EXAMPLE_HOSTED.get(user_input.app_instance.app_name, {})
        is_demo_ds = user_input.dataset_name in valid_app
        if is_demo_ds:
            config[
                'CUSTOM_DNS'
            ] = f'now-example-{user_input.app_instance.app_name}-{user_input.dataset_name}.dev.jina.ai'
            config['CUSTOM_DNS'] = config['CUSTOM_DNS'].replace('_', '-')

    return config


def common_setup(
    app_instance: JinaNOWApp,
    user_input: UserInput,
    dataset: DocumentArray,
    encoder_uses: str,
    encoder_uses_with: Dict,
    indexer_uses: str,
    pre_trained_embedding_size: int,
    kubectl_path: str,
    encoder_with: Optional[Dict] = {},
    indexer_resources: Optional[Dict] = {},
) -> Dict:
    # should receive pre embedding size
    finetune_settings = parse_finetune_settings(
        pre_trained_embedding_size=pre_trained_embedding_size,
        user_input=user_input,
        dataset=dataset,
        finetune_datasets=app_instance.finetune_datasets,
        model_name='mlp',
        add_embeddings=True,
        loss='TripletMarginLoss',
    )
    tags = _extract_tags_for_indexer(deepcopy(dataset[0]), user_input)
    env_dict = common_get_flow_env_dict(
        finetune_settings=finetune_settings,
        encoder_uses=encoder_uses,
        encoder_with=encoder_with,
        encoder_uses_with=encoder_uses_with,
        pre_trained_embedding_size=pre_trained_embedding_size,
        indexer_uses=indexer_uses,
        indexer_resources=indexer_resources,
        user_input=user_input,
        tags=tags,
    )

    if finetune_settings.perform_finetuning:
        try:
            artifact_id, token = finetune(
                finetune_settings=finetune_settings,
                app_instance=app_instance,
                dataset=dataset,
                user_input=user_input,
                env_dict=env_dict,
                kubectl_path=kubectl_path,
            )

            finetune_settings.finetuned_model_artifact = artifact_id
            finetune_settings.token = token

            env_dict['FINETUNE_ARTIFACT'] = finetune_settings.finetuned_model_artifact
            env_dict['JINA_TOKEN'] = finetune_settings.token
        except Exception as e:
            print(
                'Finetuning is currently offline. The program execution still continues without'
                ' finetuning. Please report the following exception to us:'
            )
            import traceback

            traceback.print_exc()
            finetune_settings.perform_finetuning = False

    app_instance.set_flow_yaml(
        finetuning=finetune_settings.perform_finetuning, dataset_len=len(dataset)
    )
    return env_dict


def get_email():
    try:
        with open(user('~/.jina/config.json')) as fp:
            config_val = json.load(fp)
            user_token = config_val['auth_token']
            client = hubble.Client(token=user_token, max_retries=None, jsonify=True)
            response = client.get_user_info()
        if 'email' in response['data']:
            return response['data']['email']
        return ''
    except FileNotFoundError:
        return ''


def get_indexer_config(
    num_indexed_samples: int,
    elastic: Optional[bool] = False,
    kubectl_path: str = None,
    deployment_type: str = None,
) -> Dict:
    """Depending on the number of samples, which will be indexed, indexer and its resources are determined.

    :param num_indexed_samples: number of samples which will be indexed; should incl. chunks for e.g. text-to-video app
    :param elastic: hack to use NOWElasticIndexer, should be changed in future.
    :param kubectl_path: path to kubectl binary
    :param deployment_type: deployment type, e.g. 'remote' or 'local'
    :return: dict with indexer and its resource config
    """

    if elastic and deployment_type == 'local':
        config = {
            'indexer_uses': f'{name_to_id_map.get("NOWElasticIndexer")}/{NOW_ELASTIC_INDEXER_VERSION}',
            'hosts': setup_elastic_service(kubectl_path),
        }
    elif elastic and deployment_type == 'remote':
        raise ValueError(
            'NOWElasticIndexer is currently not supported for remote deployment. Please use local deployment.'
        )
    else:
        config = {
            'indexer_uses': f'{name_to_id_map.get("NOWQdrantIndexer16")}/{NOW_QDRANT_INDEXER_VERSION}'
        }
    threshold1 = 250_000
    if num_indexed_samples <= threshold1:
        config['indexer_resources'] = {'INDEXER_CPU': 0.1, 'INDEXER_MEM': '2G'}
    else:
        config['indexer_resources'] = {'INDEXER_CPU': 1.0, 'INDEXER_MEM': '4G'}

    return config


def _extract_tags_for_indexer(d: Document, user_input):
    with tempfile.TemporaryDirectory() as tmpdir:
        if user_input and user_input.dataset_type == DatasetTypes.S3_BUCKET:
            maybe_download_from_s3(
                docs=DocumentArray([d]),
                tmpdir=tmpdir,
                user_input=user_input,
                max_workers=1,
            )
    tags = set()
    for tag, _ in d.tags.items():
        tags.add((tag, str(tag.__class__.__name__)))
    final_tags = [list(tag) for tag in tags]
    if user_input.app_instance.output_modality in [
        Modalities.IMAGE,
        Modalities.VIDEO,
    ]:
        final_tags.append([TAG_INDEXER_DOC_HAS_TEXT, str(bool.__name__)])
    return final_tags


def setup_elastic_service(
    kubectl_path: str,
) -> str:
    """Setup ElasticSearch service and return a connection string to connect to the service with.

    :param kubectl_path: path to kubectl binary
    :return: connection string for connecting to the ElasticSearch service.
    """
    cur_dir = pathlib.Path(__file__).parent.resolve()
    cmd(
        f'{kubectl_path} create -f https://download.elastic.co/downloads/eck/2.4.0/crds.yaml'
    )
    cmd(
        f'{kubectl_path} apply -f https://download.elastic.co/downloads/eck/2.4.0/operator.yaml'
    )
    cmd(f'{kubectl_path} create ns nowapi')
    cmd(f'{kubectl_path} apply -f {cur_dir}/../deployment/elastic_kind.yml')
    num_retries = 0
    es_password, error_msg = '', b''
    while num_retries < MAX_RETRIES:
        es_password, error_msg = cmd(
            [
                kubectl_path,
                "get",
                "secret",
                "quickstart-es-elastic-user",
                "-o",
                "go-template='{{.data.elastic | base64decode}}'",
            ]
        )
        if es_password:
            es_password = es_password.decode("utf-8")[1:-1]
            break
        else:
            num_retries += 1
            time.sleep(2)
    if not es_password:
        raise Exception(error_msg.decode("utf-8"))
    host = f"https://elastic:{es_password}@quickstart-es-http.default:9200"
    return host


def _get_clip_apps_with_dict(user_input: UserInput) -> Tuple[Dict, Dict]:
    """Depending on whether this app will be remotely deployed, this function returns the with
    dictionary for the CLIP executor."""
    is_remote = user_input.deployment_type == 'remote'
    encoder_with = {
        'ENCODER_HOST': EXTERNAL_CLIP_HOST if is_remote else '0.0.0.0',
        'ENCODER_PORT': 443 if is_remote else random_port(),
        'IS_REMOTE_DEPLOYMENT': is_remote,
    }
    return encoder_with



from typing import List, Optional

from pydantic import Field

from deployment.bff.app.v1.models.helper import (
    BaseIndexRequestModel,
    BaseSearchRequestModel,
    BaseSearchResponseModel,
)


# Request Model
class NowImageIndexRequestModel(BaseIndexRequestModel):
    images: Optional[List[str]] = Field(
        default=...,
        description='List of Images to index. Images should be base64encoded in `utf-8` format',
    )


class NowImageSearchRequestModel(BaseSearchRequestModel):
    image: Optional[str] = Field(
        default=None,
        description='Image query. Image should be base64encoded in `utf-8` format',
    )


# Response Model
class NowImageResponseModel(BaseSearchResponseModel):
    blob: Optional[str] = Field(
        description='Base64 encoded image in `utf-8` str format'
    )


NowImageResponseModel.update_forward_refs()



import json
import os
from typing import Dict, List, Optional

import hubble
from docarray import Document, DocumentArray
from jina import __version__ as jina_version

from now.app.base.app import JinaNOWApp
from now.common.utils import get_indexer_config
from now.constants import (
    EXECUTOR_PREFIX,
    NOW_AUTOCOMPLETE_VERSION,
    NOW_PREPROCESSOR_VERSION,
    Apps,
    Modalities,
    ModelDimensions,
    ModelNames,
)
from now.executor.name_to_id_map import name_to_id_map
from now.finetuning.data_builder import DataBuilder
from now.finetuning.run_finetuning import finetune
from now.finetuning.settings import FinetuneSettings, parse_finetune_settings
from now.now_dataclasses import Task, UserInput


class TextToTextAndImage(JinaNOWApp):
    """
    Hybrid text to text+image search combining symbolic and neural IR approaches.
    """

    def __init__(self):
        super().__init__()

    @property
    def app_name(self) -> Apps:
        return Apps.TEXT_TO_TEXT_AND_IMAGE

    @property
    def is_enabled(self) -> bool:
        return True

    @property
    def description(self) -> str:
        return (
            'Text to text+image search app combining symbolic and neural IR approaches.'
        )

    def set_flow_yaml(self, **kwargs):
        """configure the flow yaml in the Jina NOW app."""
        finetuning = kwargs.get('finetuning', False)

        flow_dir = os.path.abspath(os.path.join(__file__, '..'))

        if finetuning:
            self.flow_yaml = os.path.join(flow_dir, 'flow.yml')
        else:
            self.flow_yaml = os.path.join(flow_dir, 'pretrained_flow.yml')

    @property
    def input_modality(self) -> List[Modalities]:
        return [Modalities.TEXT]

    @property
    def output_modality(self) -> List[Modalities]:
        return [Modalities.TEXT_AND_IMAGE]

    @staticmethod
    def _create_task_config(user_input: UserInput, data_example: Document) -> Task:
        """
        Read task configuration template and replace field names.
        """
        # read task config template
        template_path = os.path.join(
            os.path.dirname(os.path.realpath(__file__)), 'task_config.json'
        )
        with open(template_path) as f:
            config_dict = json.load(f)
            config = Task(**config_dict)
        search_chunks = (
            [
                chunk
                for chunk in data_example.chunks
                if chunk.tags['field_name'] in user_input.search_fields
            ]
            if user_input.search_fields
            else data_example.chunks
        )
        text_fields = [
            chunk.tags['field_name']
            for chunk in search_chunks
            if chunk.modality == 'text'
        ]
        image_fields = [
            chunk.tags['field_name']
            for chunk in search_chunks
            if chunk.modality == 'image'
        ]
        # put field names into generation function configurations
        for encoder in config.encoders:
            if encoder.name == 'text_encoder':
                for method in encoder.training_data_generation_methods:
                    method.query.scope = text_fields
                    method.target.scope = text_fields
            elif encoder.name == 'vision_encoder':
                for method in encoder.training_data_generation_methods:
                    method.query.scope = text_fields
                    method.target.scope = image_fields
        # specify text and image field for the indexer
        config.indexer_scope['text'] = text_fields[0]
        config.indexer_scope['image'] = image_fields[0]

        user_input.task_config = config
        return user_input.task_config

    @hubble.login_required
    def setup(
        self, dataset: DocumentArray, user_input: UserInput, kubectl_path
    ) -> Dict:
        task_config = self._create_task_config(user_input, dataset[0])
        data = DataBuilder(dataset=dataset, config=task_config).build()
        env_dict = {'N_DIM': []}
        for encoder_data, encoder_type in data:
            finetune_settings = self._construct_finetune_settings(
                user_input=user_input,
                dataset=dataset,
                encoder_type=encoder_type,
            )
            if not finetune_settings.perform_finetuning:
                break
            artifact_id, token = finetune(
                finetune_settings=finetune_settings,
                app_instance=self,
                dataset=encoder_data,
                user_input=user_input,
                env_dict={},
                kubectl_path=kubectl_path,
            )

            env_dict['JINA_TOKEN'] = token
            if finetune_settings.model_name == ModelNames.CLIP:
                env_dict['CLIP_ARTIFACT'] = artifact_id
                env_dict['N_DIM'].append(ModelDimensions.CLIP)
            elif finetune_settings.model_name == ModelNames.SBERT:
                env_dict['SBERT_ARTIFACT'] = artifact_id
                env_dict['N_DIM'].append(ModelDimensions.SBERT)
            else:
                print(f'{self.app_name} only expects CLIP or SBERT models.')
                raise

        env_dict[
            'PREPROCESSOR_NAME'
        ] = f'{EXECUTOR_PREFIX}{name_to_id_map.get("NOWPreprocessor")}/{NOW_PREPROCESSOR_VERSION}'
        env_dict['APP'] = self.app_name
        indexer_config = get_indexer_config(
            len(dataset),
            elastic=True,
            kubectl_path=kubectl_path,
            deployment_type=user_input.deployment_type,
        )
        env_dict['HOSTS'] = indexer_config.get('hosts')
        env_dict['INDEXER_NAME'] = f"{EXECUTOR_PREFIX}{indexer_config['indexer_uses']}"
        env_dict['INDEXER_MEM'] = indexer_config.get('indexer_resources', {}).get(
            'INDEXER_MEM'
        )
        env_dict['JINA_VERSION'] = jina_version
        env_dict['ENCODER_NAME'] = f"{EXECUTOR_PREFIX}FinetunerExecutor/v0.9.2"
        env_dict['N_DIM'] = [ModelDimensions.SBERT, ModelDimensions.CLIP]
        env_dict['ADMIN_EMAILS'] = (
            user_input.admin_emails or [] if user_input.secured else []
        )
        env_dict['USER_EMAILS'] = (
            user_input.user_emails or [] if user_input.secured else []
        )
        env_dict[
            'AUTOCOMPLETE_EXECUTOR_NAME'
        ] = f'jinahub+docker://NOWAutoCompleteExecutor2/{NOW_AUTOCOMPLETE_VERSION}'

        env_dict['API_KEY'] = (
            [user_input.api_key] if user_input.secured and user_input.api_key else []
        )
        self.set_flow_yaml()
        super().setup(dataset, user_input, kubectl_path)
        return env_dict

    def _construct_finetune_settings(
        self,
        user_input: UserInput,
        dataset: DocumentArray,
        encoder_type: str,
    ) -> FinetuneSettings:
        finetune_settings = parse_finetune_settings(
            user_input=user_input,
            dataset=dataset,
            model_name=self._model_name(encoder_type),
            loss=self._loss_function(encoder_type),
            add_embeddings=False,
        )
        # temporary adjustments to work with small text+image dataset
        finetune_settings.epochs = 2
        finetune_settings.num_val_queries = 5
        finetune_settings.train_val_split_ration = 0.75
        return finetune_settings

    @staticmethod
    def _model_name(encoder_type: Optional[str] = None) -> str:
        """Name of the model used in case of fine-tuning."""
        if encoder_type == 'text-to-text':
            return ModelNames.SBERT
        elif encoder_type == 'text-to-image':
            return ModelNames.CLIP

    @staticmethod
    def _loss_function(encoder_type: Optional[str] = None) -> str:
        """Loss function used during fine-tuning."""
        if encoder_type == 'text-to-image':
            return 'CLIPLoss'
        return 'TripletMarginLoss'



import os
import subprocess
from time import sleep

import yaml
from jina import DocumentArray

from now.executor.abstract.base_indexer import NOWBaseIndexer as Executor


class NOWQdrantIndexer16(Executor):
    """NOWQdrantIndexer16 indexes Documents into a Qdrant server using DocumentArray  with `storage='qdrant'`."""

    # override
    def construct(self, **kwargs):
        setup_qdrant_server(self.workspace, self.logger)
        self._index = DocumentArray(
            storage='qdrant',
            config={
                'collection_name': 'Persisted',
                'host': 'localhost',
                'port': 6333,
                'n_dim': self.dim,
                'distance': self.metric,
                'ef_construct': None,
                'm': None,
                'scroll_batch_size': 64,
                'full_scan_threshold': None,
                'serialize_config': {},
                'columns': self.columns,
            },
        )
        self.range_operators = ['$gt', '$lt', '$get', '$let']

    # override
    def batch_iterator(self):
        """Iterator which iterates through the documents of self._index and yields batches"""
        batch = []
        for item in self._index:
            batch.append(item)
            if len(batch) == 1000:
                yield batch
                batch = []
        if batch:
            yield batch

    # override
    def convert_filter_syntax(self, search_filter={}, search_filter_not={}):
        """Supports exact matches and range filter."""

        def _convert_filter(filter_dict):
            conditions = []
            for attribute, condition in filter_dict.items():
                for operator, value in condition.items():
                    if operator in self.range_operators:
                        operator_type = 'range'
                        operator_string = operator.replace('$', '')
                    elif operator in ['$eq', '$regex', '$in']:
                        operator_type = 'match'
                        operator_string = 'value'
                    else:
                        continue
                    conditions.append(
                        {"key": attribute, operator_type: {operator_string: value}}
                    )
            return conditions

        search_filter_ret = {}
        if search_filter:
            search_filter_ret['must'] = _convert_filter(search_filter)
        if search_filter_not:
            search_filter_ret['must_not'] = _convert_filter(search_filter_not)
        return search_filter_ret

    # override
    def index(self, docs: DocumentArray, parameters: dict, **kwargs):
        """Index new documents"""
        # qdrant needs a list of values when filtering on sentences
        for d in docs:
            if 'title' in d.tags:
                d.tags['title'] = d.tags['title'].lower().split()
        self._index.extend(docs)

    # override
    def delete(self, documents_to_delete, parameters: dict = {}, **kwargs):
        """
        Delete endpoint to delete document/documents from the index.
        Filter conditions can be passed to select documents for deletion.
        """
        for d in documents_to_delete:
            del self._index[d.id]

    # override
    def search(
        self,
        docs: DocumentArray,
        parameters: dict,
        limit: int,
        search_filter: dict,
        **kwargs,
    ):
        """Perform a vector similarity search and retrieve `Document` matches."""
        docs.match(self._index, filter=search_filter, limit=limit)


def setup_qdrant_server(workspace, logger):
    qdrant_config_path = '/qdrant/config/production.yaml'
    if workspace and os.path.exists(qdrant_config_path):
        logger.info('set new storage to network file system location in WOLF')
        qdrant_config = yaml.safe_load(open(qdrant_config_path))

        qdrant_config['storage'] = {
            'storage_path': os.path.join(workspace, 'user_input.json')
        }
        yaml.safe_dump(qdrant_config, open(qdrant_config_path, 'w'))
        logger.info('if qdrant exists, then start it')
    try:
        subprocess.Popen(['./run-qdrant.sh'])
        sleep(3)
        logger.info('Qdrant server started')
    except FileNotFoundError:
        logger.info('Qdrant not found, locally. So it won\'t be started.')
